#!/usr/bin/ruby

# Given a Visual Studio project, create a CMakeLists.txt file which optionally
# allows for ongoing side-by-side operation (e.g. on Linux, Mac)
# together with the existing static .vcproj project on the Windows side.
# Provides good support for simple DLL/Static/Executable projects,
# but custom build steps and build events are ignored.

# Author: Jesper Eskilson
# Email: jesper [at] eskilson [dot] se
# Version: 1.0
# Large list of extensions by Andreas Mohr andi [at] lisas [period] de
# (list _all_ configuration types, add indenting, add per-platform configuration
# of definitions, dependencies and includes, add optional includes
# to provide static content, thus allowing for a nice on-the-fly
# generation mode of operation _side-by-side_ existing and _updated_ .vcproj files,
# fully support recursive handling of all .vcproj file groups (filters))

# If you add code, please try to keep this file generic and modular, to enable other people
# to hook into a particular part easily and thus keep your local additions _separate_.

# Always make sure that a simple vcproj2cmake.rb run will result in a
# fully working _self-contained_ CMakeLists.txt, no matter how small
# the current vcproj2cmake config environment is
# (i.e., it needs to work even without a single specification file)

# TODO:
# - perhaps there's a way to provide more precise/comfortable hook script handling?
# - should continue with clean separation of .vcproj content parsing and .vcproj output
#   generation (e.g. in preparation for .vcxproj support)
# - try to come up with an ingenious way to near-_automatically_ handle those pesky repeated
#   dependency requirements of several sub projects
#   (e.g. the component-based Boost Find scripts, etc.) instead of having to manually
#   write custom hook script content (which cannot be kept synchronized
#   with changes _automatically_!!) each time due to changing components and libraries.

require 'fileutils'
require 'rexml/document'
include FileUtils::Verbose

# Usage: vcproj2cmake.rb <input.vcproj> [<output CMakeLists.txt>] [<master project directory>]

filename = ARGV.shift
output_file = ARGV.shift or output_file = File.join(File.dirname(filename), "CMakeLists.txt")
master_project_dir = ARGV.shift or nil

# local config directory as created in every project which needs specific settings
# (possibly required in root project space only)
v2c_config_dir = "./cmake/vcproj2cmake"
filename_map_inc = "#{v2c_config_dir}/include_mappings.txt"
filename_map_def = "#{v2c_config_dir}/define_mappings.txt"
filename_map_dep = "#{v2c_config_dir}/dependency_mappings.txt"

$myindent = 0

def puts_ind(chan, str)
  chan.print ' ' * $myindent
  chan.puts str
end

# Change \ to /, and remove leading ./
def normalize(p)
  felems = p.gsub("\\", "/").split("/")
  # DON'T eradicate single '.'!!
  felems.shift if felems[0] == "." and felems.size > 1
  File.join(felems)
end

def escape_char(in_string, esc_char)
  #puts "in_string #{in_string}"
  in_string.gsub!(/#{esc_char}/, "\\#{esc_char}")
  #puts "in_string quoted #{in_string}"
end

def escape_backslash(in_string)
  # "Escaping a Backslash In Ruby's Gsub": "The reason for this is that
  # the backslash is special in the gsub method. To correctly output a
  # backslash, 4 backslashes are needed.". Oerks - oh well, do it.
  # hrmm, seems we need some more even...
  # (or could we use single quotes (''') for that? Too lazy to retry...)
  in_string.gsub!(/\\/, "\\\\\\\\")
end

def read_mappings(filename_mappings, mappings)
  # line format is: "tag:PLATFORM1:PLATFORM2=tag_replacement2:PLATFORM3=tag_replacement3"
  if File.exists?(filename_mappings)
    #Hash[*File.read(filename_mappings).scan(/^(.*)=(.*)$/).flatten]
    File.open(filename_mappings, 'r').each do |line|
      next if line =~ /^\s*#/
      b, c = line.chomp.split(/:/)
      mappings[b] = c
    end
  else
    puts "#{filename_mappings} NOT AVAILABLE"
  end
  #puts mappings["kernel32"]
  #puts mappings["mytest"]
end

def read_mappings_combined(filename_mappings, mappings)
  if $master_project_dir
    # read common mappings to be used by all sub projects
    read_mappings("#{$master_project_dir}/#{filename_mappings}", mappings)
  end
  read_mappings(filename_mappings, mappings)
end

def push_platform_def(platform_defs, platform, def_value)
  #puts "adding #{def_value} on platform #{platform}"
  if platform_defs[platform].nil?
    platform_defs[platform] = Array.new
  end
  platform_defs[platform].push(def_value)
end

def parse_platform_conversions(platform_defs, arr_defs, map_defs)
  arr_defs.each { |currdef|
    #puts map_defs[currdef]
    map_line = map_defs[currdef]
    if map_line.nil?
      # hmm, no direct match! Try to figure out whether any map entry
      # is a regex which would match our currdef
      map_defs.each do |key, value|
        if currdef =~ /^#{key}$/
          puts "KEY: #{key} currdef #{currdef}"
          map_line = value
        end
      end
    end
    if map_line.nil?
      # no mapping? --> unconditionally use this define
      push_platform_def(platform_defs, "ALL", currdef)
    else
      map_line.chomp.split(/\|/).each do |platform_element|
        #puts "platform_element #{platform_element}"
        platform, replacement_def = platform_element.split(/=/)
        if platform.empty?
          # specified a replacement without a specific platform?
          # ("tag:=REPLACEMENT")
          # --> unconditionally use it!
          platform = "ALL"
        else
          if replacement_def.nil?
            replacement_def = currdef
          end
        end
        push_platform_def(platform_defs, platform, replacement_def)
      end
    end
  }
end

def dump_build_attributes(cmake_command, element_prefix, out, arr_defs, map_defs, cmake_command_arg)
  # the container for the list of _actual_ dependencies as stated by the project
  all_platform_defs = Hash.new
  parse_platform_conversions(all_platform_defs, arr_defs, map_defs)
  all_platform_defs.each { |key, arr_platdefs|
    #puts "arr_platdefs: #{arr_platdefs}"
    next if arr_platdefs.empty?
    arr_platdefs.uniq!
    out.puts
    specific_platform = !(key.eql?("ALL"))
    if specific_platform
      puts_ind(out, "if(#{key})")
      $myindent += 2
    end
    if cmake_command_arg.nil?
      puts_ind(out, "#{cmake_command}(")
    else
      puts_ind(out, "#{cmake_command}(#{cmake_command_arg}")
    end
    arr_platdefs.each do |currdef|
      puts_ind(out, "  #{element_prefix}#{currdef}")
    end
    puts_ind(out, ")")
    if specific_platform
      $myindent -= 2
      puts_ind(out, "endif(#{key})")
    end
  }
end

def parse_file(project, file, arr_sources)
  projname = project.attributes["Name"]
  f = normalize(file.attributes["RelativePath"])

  # Ignore header files
  return if f =~ /\.(h|H|lex|y|ico|bmp|txt)$/

  # Ignore files which have the ExcludedFromBuild attribute set to TRUE
  excluded_from_build = false
  file.elements.each("FileConfiguration") { |file_config|
    #file_config.elements.each('Tool[@Name="VCCLCompilerTool"]') { |compiler|
    #  if compiler.attributes["UsePrecompiledHeader"]
    #}
    excl_build = file_config.attributes["ExcludedFromBuild"]
    if not excl_build.nil? and excl_build.downcase == "true"
      excluded_from_build = true
      return # no complex handling, just return
    end
  }

  # Ignore files with custom build steps
  included_in_build = true
  file.elements.each("FileConfiguration/Tool") { |tool|
    if tool.attributes["Name"] == "VCCustomBuildTool"
      included_in_build = false
      return # no complex handling, just return
    end
  }

  # Verbosely ignore IDL generated files
  if f =~/_(i|p).c$/
    # see file_mappings.txt comment above
    puts "#{projname}::#{f} as an IDL generated file: skipping! FIXME: should be platform-dependent."
    included_in_build = false
    return # no complex handling, just return
  end

  # Verbosely ignore .lib "sources"
  if f =~ /\.lib$/
    # probably these entries are supposed to serve as dependencies
    # (i.e., non-link header-only include dependency, to ensure
    # rebuilds in case of foreign-library header file changes).
    # Not sure whether these were added by users or
    # it's actually some standard MSVS mechanism... FIXME
    puts "#{projname}::#{f} registered as a \"source\" file!? Skipping!"
    included_in_build = false
    return # no complex handling, just return
  end

  if not excluded_from_build and included_in_build
  	  arr_sources.push(f)
  end
end

Files_str = Struct.new(:name, :arr_sub_filters, :arr_files)

def vc8_parse_file_list(project, vcproj_filter, files_str)
  file_group_name = vcproj_filter.attributes["Name"]
  if not file_group_name.nil?
    files_str[:name] = file_group_name
  else
    files_str[:name] = "COMMON"
  end
  puts "parsing files group #{files_str[:name]}"

  vcproj_filter.elements.each("Filter") { |subfilter|
    # skip file filters that have a SourceControlFiles property
    # that's set to false, i.e. files which aren't under version
    # control (such as IDL generated files).
    # This experimental check might be a little rough after all...
    # yes, FIXME: on Win32, these files likely _should_ get listed
    # after all. We should probably do a platform check in such
    # cases, i.e. add support for a file_mappings.txt
    scfiles = subfilter.attributes["SourceControlFiles"]
    if not scfiles.nil? and scfiles.downcase == "false"
      puts "#{name}: SourceControlFiles set to false, listing generated files? --> skipping!"
      next
    end
    if files_str[:arr_sub_filters].nil?
      files_str[:arr_sub_filters] = Array.new()
    end
    subfiles_str = Files_str.new()
    files_str[:arr_sub_filters].push(subfiles_str)
    vc8_parse_file_list(project, subfilter, subfiles_str)
  }

  arr_sources = Array.new()
  vcproj_filter.elements.each("File") { |file|
    parse_file(project, file, arr_sources)
  } # |file|

  if not arr_sources.empty?
    files_str[:arr_files] = arr_sources
  end
end

def write_file_list(project, files_str, parent_source_group, arr_sub_sources_for_parent, out)
  group = files_str[:name]
  if not files_str[:arr_sub_filters].nil?
    arr_sub_filters = files_str[:arr_sub_filters]
  end
  if not files_str[:arr_files].nil?
    arr_local_sources = files_str[:arr_files].clone
  end

  # TODO: cmake is said to have a weird bug in case of parent_source_group being "Source Files"
  # http://www.mail-archive.com/cmake@cmake.org/msg05002.html
  if parent_source_group.nil?
    this_source_group = ""
  else
    if parent_source_group == ""
      this_source_group = group
    else
      this_source_group = "#{parent_source_group}\\\\#{group}"
    end
  end

  # process sub-filters, have their main source variable added to arr_my_sub_sources
  arr_my_sub_sources = Array.new()
  if not arr_sub_filters.nil?
    $myindent += 2
    arr_sub_filters.each { |subfilter|
      #puts "writing: #{subfilter}"
      write_file_list(project, subfilter, this_source_group, arr_my_sub_sources, out)
    }
    $myindent -= 2
  end

  group_tag = this_source_group.clone
  group_tag.gsub!(' ','_')
  group_tag.gsub!('\\','_')

  # process our hierarchy's own files
  if not arr_local_sources.nil?
    source_files_variable = "SOURCES_files_#{group_tag}"
    out.puts
    puts_ind(out, "set(#{source_files_variable}" )
    arr_local_sources.each { |source|
      puts "quotes now: #{source}"
      if source.include? ' '
        # quote arguments with spaces (yes, I don't know how to
        # preserve quotes properly other than to actually open-code it
        # in such an ugly way - ARGH!!)
        puts_ind(out, "  " + "\"#{source}\"")
      else
        puts_ind(out, "  " + "#{source}")
      end
    }
    puts_ind(out, ")")
    # create source_group() of our local files
    if not parent_source_group.nil?
      puts_ind(out, "source_group(\"#{this_source_group}\" FILES ${#{source_files_variable}})")
    end
  end
  if not source_files_variable.nil? or not arr_my_sub_sources.empty?
    sources_variable = "SOURCES_#{group_tag}";
    out.puts
    puts_ind(out, "set(SOURCES_#{group_tag}")
    $myindent += 2;
    # dump sub filters...
    arr_my_sub_sources.each { |source|
      puts_ind(out, "${#{source}}")
    }
    # ...then our own files
    if not source_files_variable.nil?
      puts_ind(out, "${#{source_files_variable}}")
    end
    $myindent -= 2;
    puts_ind(out, ")")
    # add our source list variable to parent return
    arr_sub_sources_for_parent.push(sources_variable)
  end
end

################
#     MAIN     #
################

if File.exists?(output_file)
  mv(output_file, output_file + ".backup")
end

File.open(output_file, "w") { |out|

  out.puts "#"
  out.puts "# TEMPORARY Build file, AUTO-GENERATED by vcproj2cmake.rb"
  out.puts "# DO NOT CHECK INTO VERSION CONTROL OR APPLY \"PERMANENT\" MODIFICATIONS!!"
  out.puts "#"
  out.puts
  # bogus version line to make cmake happy.
  # Not sure whether the commands in here are really all 2.4 compatible.
  out.puts "cmake_minimum_required(VERSION 2.4)"

  out.puts "if(COMMAND cmake_policy)"
  # manual quoting of brackets in definitions doesn't seem to work otherwise,
  # in cmake 2.6.4-7.el5 with CMP0005 OLD.
  out.puts "  if(POLICY CMP0005)"
  out.puts "    cmake_policy(SET CMP0005 NEW) # automatic quoting of brackets"
  out.puts "  endif(POLICY CMP0005)"
  out.puts
  out.puts("  # we do want the includer to be affected by our updates,")
  out.puts("  # since it might define project-global settings.")
  out.puts("  if(POLICY CMP0011)")
  out.puts("    cmake_policy(SET CMP0011 OLD)")
  out.puts("  endif(POLICY CMP0011)")
  out.puts "endif(COMMAND cmake_policy)"

  File.open(filename) { |io|
    doc = REXML::Document.new io

    # try to point to cmake/Modules of the topmost directory of the vcproj2cmake conversion tree.
    if master_project_dir.nil?
      module_path_element = "${PROJECT_SOURCE_DIR}/../../cmake/Modules"
    else
      module_path_element = "#{master_project_dir}/cmake/Modules"
    end

    # NOTE: could be problematic to _append_ new paths, should perhaps prepend them
    # (otherwise not able to provide proper overrides)
    out.puts
    puts_ind(out, "list(APPEND CMAKE_MODULE_PATH #{module_path_element})")

    # "export" our internal v2c_config_dir variable (to be able to reference it in CMake scripts as well)
    out.puts
    puts_ind(out, "set(V2C_CONFIG_DIR #{v2c_config_dir})")

    # this CMakeLists.txt-global optional include could be used e.g.
    # to skip the entire build of this file on certain platforms:
    # if(PLATFORM) message(STATUS "not supported") return() ...
    # (note that we appended CMAKE_MODULE_PATH _prior_ to this include()!)
    out.puts
    puts_ind(out, "include(${V2C_CONFIG_DIR}/hook_pre.txt OPTIONAL)")

    doc.elements.each("VisualStudioProject") { |project|

      name = project.attributes["Name"]

      out.puts
      # TODO: figure out language type (C CXX etc.) and add it to project() command
      puts_ind(out, "project( #{name} )")

      ## sub projects will inherit, and we _don't_ want that...
      # DISABLED: now to be done by MasterProjectDefaults_vcproj2cmake module if needed
      #puts_ind(out, "# reset project-local variables")
      #puts_ind(out, "set( V2C_LIBS )")
      #puts_ind(out, "set( V2C_SOURCES )")

      out.puts
      out.puts "# this part is for including a file which contains"
      out.puts "# _globally_ applicable settings for all sub projects of a master project"
      out.puts "# (compiler flags, path settings, platform stuff, ...)"
      out.puts "# e.g. have vcproj2cmake-specific MasterProjectDefaults_vcproj2cmake"
      out.puts "# which then _also_ includes a global MasterProjectDefaults module"
      out.puts "# for _all_ CMakeLists.txt. This needs to sit post-project()"
      out.puts "# since e.g. compiler info is dependent on a valid project."
      puts_ind(out, "# MasterProjectDefaults_vcproj2cmake is supposed to define")
      puts_ind(out, "# generic settings (such as V2C_HOOK_PROJECT, defined as e.g.")
      puts_ind(out, "# ./cmake/vcproj2cmake/hook_project.txt, and other hook include variables below).")
      puts_ind(out, "# NOTE: it usually should also reset variables V2C_LIBS, V2C_SOURCES etc.")
      puts_ind(out, "# as used below since they should contain directory-specific contents only, not accumulate!")
      puts_ind(out, "include(MasterProjectDefaults_vcproj2cmake OPTIONAL)")

      puts_ind(out, "# hook e.g. for invoking Find scripts as expected by")
      puts_ind(out, "# the _LIBRARIES / _INCLUDE_DIRS mappings created")
      puts_ind(out, "# by your include/dependency map files.")
      puts_ind(out, "include(${V2C_HOOK_PROJECT} OPTIONAL)")

      main_files = Files_str.new()
      project.elements.each("Files") { |files|
      	vc8_parse_file_list(project, files, main_files)
      }
      arr_sub_sources = Array.new()
      $myindent += 2
      write_file_list(project, main_files, nil, arr_sub_sources, out)
      $myindent -= 2

      if not arr_sub_sources.empty?
        # add a ${V2C_SOURCES} variable to the list, to be able to append
        # all sorts of (auto-generated, ...) files to this list within
        # hook includes, _right before_ creating the target with its sources.
        arr_sub_sources.push("V2C_SOURCES")
      else
        puts "WARNING: #{name}: no source files at all!? (header-based project?)"
      end
      puts_ind(out, "include(${V2C_HOOK_POST_SOURCES} OPTIONAL)")

      # TODO: should we offer a hook_post_sources.txt here?

      project.elements.each("Configurations/Configuration") { |config|
        config_name = config.attributes["Name"].split("|")[0]
        out.puts
        out.puts "if(CMAKE_BUILD_TYPE MATCHES \"#{config_name}\")"
        $myindent += 2

        # FIXME: do we need to actively reset CMAKE_MFC_FLAG / CMAKE_ATL_FLAG
        # (i.e. best also set it in case of 0?), since subdirectories shouldn't inherit?

        # 0 == no MFC
        # 1 == static MFC
        # 2 == shared MFC
        use_of_mfc = config.attributes["UseOfMFC"].to_i
        if use_of_mfc > 0
          out.puts
          puts_ind(out, "set(CMAKE_MFC_FLAG #{use_of_mfc})")
        end

        # ok, there's no CMAKE_ATL_FLAG yet, AFAIK, but still prepare
        # for it (also to let people probe on this in hook includes)
        use_of_atl = config.attributes["UseOfATL"].to_i
        if use_of_atl > 0
          out.puts
          puts_ind(out, "set(CMAKE_ATL_FLAG #{use_of_atl})")
        end

        config.elements.each('Tool[@Name="VCCLCompilerTool"]') { |compiler|

          if compiler.attributes["AdditionalIncludeDirectories"]

            arr_includes = Array.new()

            include_dirs = compiler.attributes["AdditionalIncludeDirectories"].split(/[,;]/).sort.each { |s|
                incpath = normalize(s).strip
                #puts "include is '#{incpath}'"
                if incpath["Vc7/atlmfc/src/mfc"]
        	  incpath = "${MFC_INCLUDE}";
        	end
                arr_includes.push(incpath)
            }
            map_includes = Hash.new()
            read_mappings_combined(filename_map_inc, map_includes)
            dump_build_attributes("include_directories", "", out, arr_includes, map_includes, nil)
          end
          arr_defines = Array.new()
          if compiler.attributes["PreprocessorDefinitions"]

            compiler.attributes["PreprocessorDefinitions"].split(";").sort.each { |s|
            str_define, str_setting = s.strip.split(/=/)
            if str_setting.nil?
                    arr_defines.push(str_define)
            else
              if str_setting =~ /[\(\)]+/
                escape_char(str_setting, '\\(')
                escape_char(str_setting, '\\)')
              end
                    arr_defines.push("#{str_define}=#{str_setting}")
            end
            }

          end

          if use_of_mfc == 2
            arr_defines.push("_AFXEXT")
            arr_defines.push("_AFXDLL")
          end

          map_defines = Hash.new()
          read_mappings_combined(filename_map_def, map_defines)
          # FIXME: investigate whether we could instead properly restrict the
          # definitions to the target itself, via set_property(TARGET
          # COMPILE_DEFINITIONS)! Well, COMPILE_DEFINITIONS is
          # problematic: it is _not_ an open-coded list of flags, it is
          # pure definitions which will then get appended with -D each.
          # Dito for includes etc.?
          dump_build_attributes("add_definitions", "-D", out, arr_defines, map_defines, nil)
        }

        config_type = config.attributes["ConfigurationType"].to_i

        out.puts
        puts_ind(out, "# hook include after all definitions have been made")
        puts_ind(out, "# (but _before_ target is created using the source list!)")
        puts_ind(out, "include(${V2C_HOOK_POST_DEFINITIONS} OPTIONAL)")

        # create a target only in case we do have any content at all
        #if not main_files[:arr_sub_filters].empty? or not main_failes[:arr_files].empty?
        if not arr_sub_sources.empty?

          # first add source reference, then create target

          out.puts
          puts_ind(out, "set(SOURCES")
          $myindent += 2
          arr_sub_sources.each { |group_tag|
            puts_ind(out, "${#{group_tag}}")
          }
          $myindent -= 2
          puts_ind(out, ")")

          # FIXME: should use a macro like rosbuild_add_executable(),
          # http://www.ros.org/wiki/rosbuild/CMakeLists ,
          # https://kermit.cse.wustl.edu/project/robotics/browser/trunk/vendor/ros/core/rosbuild/rosbuild.cmake?rev=3
          # to be able to detect non-C++ file types within a source file list
          # and add a hook to handle them specially.
          target = nil
          if config_type == 1    # Executable
            target = name
            #puts_ind(out, "add_executable_vcproj2cmake( #{name} WIN32 ${SOURCES} )")
            # TODO: perhaps for real cross-platform binaries (i.e.
            # console apps not needing a WinMain()), we should detect
            # this and not use WIN32 in this case...
            puts_ind(out, "add_executable( #{target} WIN32 ${SOURCES} )")
          elsif config_type == 2    # DLL
            target = name
            #puts_ind(out, "add_library_vcproj2cmake( #{name} SHARED ${SOURCES} )")
            puts_ind(out, "add_library( #{target} SHARED ${SOURCES} )")
          elsif config_type == 4    # Static
            target = name
            #puts_ind(out, "add_library_vcproj2cmake( #{name} STATIC ${SOURCES} )")
            puts_ind(out, "add_library( #{target} STATIC ${SOURCES} )")
          elsif config_type == 0 # seems to be some sort of non-build logical collection project, TODO: investigate!
          elsif
            $stderr.puts "Project type #{config_type} not supported."
            exit 1
          end

          if not target.nil?
            arr_dependencies = Array.new()
            config.elements.each('Tool[@Name="VCLinkerTool"]') { |linker|
              deps = linker.attributes["AdditionalDependencies"]
              if deps and deps.length > 0
                deps.split.each  { |lib|
                  lib = lib.gsub(/\\/, "/")
                  arr_dependencies.push(File.basename(lib, ".lib"))
                }
              end
            }
            map_dependencies = Hash.new()
            read_mappings_combined(filename_map_dep, map_dependencies)
            # TODO: we probably should reset V2C_LIBS at the beginning of every CMakeLists.txt
            # (see ldd -u -r for superfluous link parts)
            arr_dependencies.push("${V2C_LIBS}")
            dump_build_attributes("target_link_libraries", "", out, arr_dependencies, map_dependencies, name)

            # make sure to keep CMake VS_KEYWORD property in our conversion, too...
            if not project.attributes["Keyword"].nil?
              keyword = project.attributes["Keyword"]
              out.puts
              puts_ind(out, "set_property(TARGET #{target} PROPERTY VS_KEYWORD #{keyword})")
            end

            # keep source control integration in our conversion!
            # FIXME: does it really work? Then reply to
            # http://public.kitware.com/mantis/view.php?id=10237 !!
            if not project.attributes["SccProjectName"].nil?
              scc_project_name = project.attributes["SccProjectName"].clone
              # hmm, perhaps need to use CGI.escape since chars other than just '"' might need to be escaped?
              # NOTE: needed to clone() this string above since otherwise modifying (same) source object!!
              escape_char(scc_project_name, '"')
              scc_local_path = project.attributes["SccLocalPath"].clone
              scc_provider = project.attributes["SccProvider"].clone
              out.puts
              puts_ind(out, "set_property(TARGET #{target} PROPERTY VS_SCC_PROJECTNAME \"#{scc_project_name}\")")
              if scc_local_path
                escape_backslash(scc_local_path)
                escape_char(scc_local_path, '"')
                puts_ind(out, "set_property(TARGET #{target} PROPERTY VS_SCC_LOCALPATH \"#{scc_local_path}\")")
              end
              if scc_provider
                escape_char(scc_provider, '"')
                puts_ind(out, "set_property(TARGET #{target} PROPERTY VS_SCC_PROVIDER \"#{scc_provider}\")")
              end
            end
          end # not target.nil?
        end # not arr_sources.empty?

        out.puts
        puts_ind(out, "# e.g. to be used for modifying target properties etc.")
        puts_ind(out, "include(${V2C_HOOK_POST_TARGET} OPTIONAL)")

        $myindent -= 2
        out.puts "endif(CMAKE_BUILD_TYPE MATCHES \"#{config_name}\")"
      }
    }
    out.puts
    puts_ind(out, "include(${V2C_HOOK_POST} OPTIONAL)")
  }
}

puts "Wrote #{output_file}"

